unit SDPartitionImage_FreeOTFE;

interface

uses
  Classes, Windows,
  OTFEFreeOTFEDLL_U,
  SDPartitionImage;

type
  TSDPartitionImage_FreeOTFE = class(TSDPartitionImage)
  protected
    FFilename: string;
    FMountedAs: char;
    FBytesPerSector: integer;
  public
    FreeOTFEObj: TOTFEFreeOTFEDLL;

    constructor Create(); override;
    destructor  Destroy(); override;

    function  DoMount(): boolean; override;
    procedure DoDismount(); override;

    function  ReadSector(sectorID: uint64; sector: TStream; maxSize: integer = -1): boolean; override;
    function  WriteSector(sectorID: uint64; sector: TStream; maxSize: integer = -1): boolean; override;
  published
    property  Filename: string read FFilename write FFilename;
  end;

implementation

uses
  Math,
  SDUClasses,
  SDUGeneral;

constructor TSDPartitionImage_FreeOTFE.Create();
begin
  inherited;
  FMountedAs := #0;
end;

destructor TSDPartitionImage_FreeOTFE.Destroy();
begin
  //lplp - TO IMPLEMENT
  inherited;
end;

function TSDPartitionImage_FreeOTFE.DoMount(): boolean;
var
  diskGeometry: TSDUDiskGeometry;
  retval: boolean;
begin
  retval := FALSE;
  
  FMountedAs := FreeOTFEObj.Mount(Filename);
  if (FMountedAs = #0) then
    begin
    FBytesPerSector := 0;
    end
  else
    begin
    if FreeOTFEObj.GetDiskGeometry(FMountedAs, diskGeometry) then
      begin
      FBytesPerSector := diskGeometry.BytesPerSector;
      retval := TRUE;
      end
    else
      begin
      // Dismount, but otherwise do nothing; retval already set to FALSE
      DoDismount();
      end;
    end;

  Result := retval;
end;

procedure TSDPartitionImage_FreeOTFE.DoDismount();
begin
  FreeOTFEObj.Dismount(FMountedAs);
end;

function TSDPartitionImage_FreeOTFE.ReadSector(sectorID: uint64; sector: TStream; maxSize: integer = -1): boolean;
var
  retval: boolean;
  partialSector: boolean;
  cntSectors: integer;
  tmpStream: TSDUMemoryStream;
begin
  // Short circuit...
  if (maxSize = 0) then
    begin
    Result := TRUE;
    exit;
    end;

  if (maxSize <= 0) then
    begin
    maxSize := FBytesPerSector;
    end;

  // If the amount of data to transfer isn't a multiple of the sector size, we
  // have to use a temp stream to hold the data 
  partialSector := ((maxSize mod FBytesPerSector) <> 0);
  cntSectors := (maxSize div FBytesPerSector);
  if partialSector then
    begin
    inc(cntSectors);
    tmpStream:= TSDUMemoryStream.Create();
    try
      retval := FreeOTFEObj.ReadData(FMountedAs, sectorID, cntSectors, tmpStream);
      tmpStream.Position := 0;
      sector.CopyFrom(tmpStream, maxSize);
    finally
      tmpStream.Free();
    end;

    end
  else
    begin
    retval := FreeOTFEObj.ReadData(FMountedAs, sectorID, cntSectors, sector);
    end;

  Result := retval;
end;

function TSDPartitionImage_FreeOTFE.WriteSector(sectorID: uint64; sector: TStream; maxSize: integer = -1): boolean;
var
  retval: boolean;
  partialSector: boolean;
  cntSectors: integer;
  tmpStream: TSDUMemoryStream;
begin
  retval := FALSE;

  // Short circuit...
  if (maxSize = 0) then
    begin
    Result := TRUE;
    exit;
    end;

  if (maxSize <= 0) then
    begin
    maxSize := FBytesPerSector;
    end;

  // If the amount of data to transfer isn't a multiple of the sector size, we
  // have to use a temp stream to hold the data
  partialSector := ((maxSize mod FBytesPerSector) <> 0);
  cntSectors := (maxSize div FBytesPerSector);
  if partialSector then
    begin
    tmpStream:= TSDUMemoryStream.Create();
    try
      inc(cntSectors);
      // In order to preserve the data previously written to the sector's slack
      // space, read in the previous data and overlay the new data onto it.
      if ReadSector(sectorID, tmpStream, (cntSectors * FBytesPerSector)) then
        begin
        tmpStream.Position := 0;
        tmpStream.CopyFrom(sector, maxSize);
        tmpStream.Position := 0;
        retval := FreeOTFEObj.WriteData(FMountedAs, sectorID, cntSectors, tmpStream);
        end;
    finally
      tmpStream.Free();
    end;

    end
  else
    begin
    retval := FreeOTFEObj.WriteData(FMountedAs, sectorID, cntSectors, sector);
    end;

  Result := retval;
end;


END.

